//  IoT本格学習キット スマートリモコン
//  7-3. Scratch3用プログラム（インターネット接続あり）

// ①ライブラリを読み込み
#include <WiFi.h>
#include <WiFiClientSecure.h>
#include <DHT.h>
#include <FS.h>
#include <SPIFFS.h>
#include <ESPAsyncWebServer.h>
#include <ArduinoJson.h>    // 「ArduinoJson by Benolt Blanchon 5.13.5」ライブラリを個別にインストール
#include <PubSubClient.h>   // 「PubSubClient by Nick O'Leary」を個別にインストール
#include "config.h"

// ②PubSubClientライブラリでのパケットサイズ、KeepAlive時間を個別で変更
#define MQTT_MAX_PACKET_SIZE 1024
#define MQTT_KEEPALIVE 15

// ③ピン配置
const byte LED_PIN = 22;   // LED緑
const byte IR_R_PIN = 36;  // リモコン受信
const byte IR_S_PIN = 23;  // リモコン送信
const byte LIGH_SNR = 39;  // 照度センサ(A18)
const byte TEMP_SNR = 21;  // 温・湿度センサ

// ④グローバル変数定義
AsyncWebServer server(80);        // Port80番
unsigned short irData[1500];
bool ledFlag = true;      // LED制御フラグ
WiFiClientSecure wifiClientMqtt;        // MQTT用
WiFiClientSecure wifiClientHttps;       // HTTPS用
PubSubClient mqttClient(mqtthost, 8883, wifiClientMqtt);

void setup(void) {
  // ⑤Serial設定
  Serial.begin(115200);
  Serial.println ( );
  // ⑤SPIFFS開始
  if (!SPIFFS.begin()) {
    Serial.println(F("SPIFFS Failed"));  // Failed mount
    delay(5000);
    ESP.restart();
  }
  // ⑥LED設定
  pinMode ( LED_PIN, OUTPUT );
  pinMode ( IR_R_PIN, INPUT );
  pinMode ( IR_S_PIN, OUTPUT );
  pinMode ( LIGH_SNR, INPUT );
  // ⑦無線Wi-Fi接続
  WiFi.config( ip, gateway, subnet, dnsserver );
  WiFi.begin ( ssid, password );
  // ⑧Wi-Fi接続処理（接続するまで無限ループ）
  Serial.print ( "Wi-Fi Connectting ." );
  while ( WiFi.status() != WL_CONNECTED ) {
    // ⑨LEDを１秒毎に点滅する
    ledFlag = !ledFlag;
    digitalWrite(LED_PIN, ledFlag);
    delay ( 1000 );
    Serial.print ( "." );
  }
  // ⑩Wi-Fi接続できたのでシリアルモニターにIPアドレス表示
  Serial.print ( "Wi-Fi Connected! IP address: " );
  Serial.println ( WiFi.localIP() );
  digitalWrite ( LED_PIN, true ); // LED点灯（Wi-Fi接続状態）
  // ⑪HTTPS用の証明書を設定
  wifiClientMqtt.setCACert(beebottle_ca_cert);
  wifiClientHttps.setCACert(ifttt_ca_cert);
  // ⑫時刻同期(NTP)設定
  configTime(9 * 3600L, 0, "ntp.nict.jp", "time.google.com", "ntp.jst.mfeed.ad.jp");
  // ⑬WebServerアクセス処理[LED制御]
  server.on("/led", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "webServer led:" + request->getParam("o")->value() );
      String setLedVal = request->getParam("o")->value();
      if (setLedVal == "ON") {
        digitalWrite ( LED_PIN, true );
      } else {
        digitalWrite ( LED_PIN, false );
      }
      // 応答処理
      request->send(200, "text/plain", "OK");
  });
  // ⑭WebServerアクセス処理[センサー情報更新]
  server.on("/update", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "webServer update" );
      // リモコン送信
      String sendMsg = makeSensorStr(1);
      // 応答処理
      request->send(200, "application/json", sendMsg);
  });
  // ⑮WebServerアクセス処理[リモコン送信]
  server.on("/send", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "webServer send:" + request->getParam("n")->value() );
      // リモコン送信
      String replyResult = "NG";
      if ( contRemocon(request->getParam("n")->value()) ){
        replyResult = "OK";
      }
      // 応答処理
      request->send(200, "text/plain", replyResult);
  });
  // ⑯WebServerアクセス処理[リモコン受信]
  server.on("/recv", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "webServer recv:" + request->getParam("n")->value() );
      // リモコン受信設定
      String replyResult = "NG";
      if ( setRemocon(request->getParam("n")->value()) ){
        replyResult = "OK";
      }
      // 応答処理
      Serial.println ( "send recv response." );
      request->send(200, "text/plain", replyResult);
  });
  // ⑰WebServerアクセス処理[IFTTT Webhooks]
  server.on("/ifttt", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "webServer ifttt:" + request->getParam("j")->value() );
      // リモコン受信設定
      String replyResult = "NG";
      if ( iftttWebhoos(request->getParam("j")->value()) ){
        replyResult = "OK";
      }
      // 応答処理
      Serial.println ( "send ifttt response." );
      request->send(200, "text/plain", replyResult);
  });
  // ⑱WebServerヘッダ設定及び起動
  DefaultHeaders::Instance().addHeader("Access-Control-Allow-Origin", "*");
  server.begin();
  Serial.println ( "Web server started" );
}

// ⑲loop処理（起動中、繰り返し処理）
void loop(void){
  // ⑳MQTT状態を確認し接続状態でなかったらMQTT接続処理
  if (!mqttClient.connected()) {
    reconnect();
  }
  // ㉑MQTT受信処理
  mqttClient.loop();
  delay(1);
}

// ㉒センサー情報取得
String makeSensorStr(byte jobType){
  String messageS = "";
  DHT dht(TEMP_SNR, DHT11);
  float g_temperature = dht.readTemperature();
  float g_humidity = dht.readHumidity();
  unsigned short g_light = analogRead(LIGH_SNR) * 100 / 4095;
  if (jobType == 1) {   // Send to Scratch
    messageS = "{\"t\":" + String(int(g_temperature)) + ",\"h\":" + String(int(g_humidity)) + ",\"l\":" + String(g_light) + "}";
  } else if (jobType == 2) {
    messageS = String(int(g_temperature)) + "_" + String(int(g_humidity)) + "_" + String(g_light);
  }
  return messageS;
}

// ㉓番号情報のみ取得し返答
String getNumStr(String message) {
  message.replace("broadcast ", "");
  message.replace("_", "");
  message.replace("\"", "");
  message.replace("SEND", "");
  message.replace("RECV", "");
  return message;
}

// ㉔MQTT接続処理
void reconnect() {
  byte tryCount = 0;
  // ㉕MQTT接続状態までループ処理
  while (!mqttClient.connected()) {
    Serial.println("Attempting MQTT connection...");
    // ㉖MQTT設定、接続処理
    String username = "token:";
    username += channelToken;
    mqttClient.connect(clientID, username.c_str(), NULL);
    delay(3000);
    if (tryCount > 100 ) {
      ESP.restart();
    }
    tryCount++;
  }
  Serial.println("MQTT connected");
  // ㉗MQTTメッセージ受信時の処理設定
  mqttClient.setCallback(callback);
  // ㉘MQTTメッセージを受信するTOPICの設定
  mqttClient.subscribe(topic);
}

// ㉙MQTTメッセージ受信時の処理
void callback(char* topic, byte* payload, unsigned int length) {
  char recvData[MQTT_MAX_PACKET_SIZE];
  snprintf(recvData, sizeof(recvData), "%s", payload);
  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.println("] ");
  Serial.println(recvData);
  // ㉚受信データのJSON形式を解析し変数へ保存
  StaticJsonBuffer<MQTT_MAX_PACKET_SIZE> jsonBuffer;
  JsonObject& jsonBuf = jsonBuffer.parseObject(recvData);
  // ㉛JSON形式解析が成功しなかったらエラー表示し終了
  if (!jsonBuf.success()) {
    Serial.println("parseObject() failed");
    return;
  }
  // ㉜受信データ（data）を保存
  const char* parsedPayload = jsonBuf["data"];
  // ㉝受信データ（data）が存在するか判定
  if (parsedPayload != NULL) {
    Serial.print("payload: ");
    Serial.println(parsedPayload);
    // ㉞受信データ（data）に送信リモコン番号
    String rMessage = String(parsedPayload);
    if(rMessage.indexOf("RECV") > -1){
      Serial.println("RECV");
      digitalWrite ( LED_PIN, false );
      setRemocon (getNumStr(rMessage));
      digitalWrite ( LED_PIN, true );
    } else if(rMessage.indexOf("SEND") > -1){
      digitalWrite ( LED_PIN, false );
      Serial.println("SEND");
      contRemocon (getNumStr(rMessage));
      digitalWrite ( LED_PIN, true );
    }
  }
}
